;; Generated package description from crux.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "crux" "0.6.0.20260315.62204" "A Collection of Ridiculously Useful eXtensions" '((emacs "26.1")) :commit "69e03917f6fd35e25b9a9dfd02df8ff3643f9227" :authors '(("Bozhidar Batsov" . "bozhidar@batsov.dev")) :maintainer '("Bozhidar Batsov" . "bozhidar@batsov.dev") :keywords '("convenience") :url "https://github.com/bbatsov/crux")
